﻿using System.Diagnostics.Contracts;
using System.Reactive.Disposables;
using System.Threading.Tasks;

namespace System.Reactive.Concurrency
{
	internal sealed class TaskPoolScheduler : LocalScheduler, ISchedulerLongRunning, ISchedulerPeriodic
	{
		private readonly TaskScheduler scheduler;

		public TaskPoolScheduler(TaskScheduler scheduler)
		{
			Contract.Requires(scheduler != null);

			this.scheduler = scheduler;
		}

		[ContractInvariantMethod]
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1822:MarkMembersAsStatic", Justification = "Required for code contracts.")]
		private void ObjectInvariant()
		{
			Contract.Invariant(scheduler != null);
		}

		private void Run<TState>(object state)
		{
			var data = (Tuple<Func<IScheduler, TState, IDisposable>, IScheduler, TState, CompositeDisposable>) state;

			Contract.Assume(data != null);
			Contract.Assume(data.Item1 != null);
			Contract.Assume(data.Item2 != null);
			Contract.Assume(data.Item4 != null);

			data.Item4.Add(data.Item1(data.Item2, data.Item3));
		}

		private void RunCancellable<TState>(object state)
		{
			var data = (Tuple<Action<TState, ICancelable>, TState, ICancelable>) state;

			Contract.Assume(data != null);
			Contract.Assume(data.Item1 != null);
			Contract.Assume(data.Item3 != null);

			data.Item1(data.Item2, data.Item3);
		}

		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Reliability", "CA2000:Dispose objects before losing scope", Justification = "Returned to caller.")]
		public override IDisposable Schedule<TState>(TState state, TimeSpan dueTime, Func<IScheduler, TState, IDisposable> action)
		{
			dueTime = Scheduler.Normalize(dueTime);

			if (dueTime == TimeSpan.Zero)
			{
				return Schedule(state, action);
			}

			var cancel = new CancellationDisposable();
			var disposables = new CompositeDisposable(cancel);

			Task.Delay(dueTime, cancel.Token)
				.ContinueWith(
					(task, s) => Run<TState>(s),
					Tuple.Create(action, this, state, disposables),
					cancel.Token,
					TaskContinuationOptions.NotOnCanceled | TaskContinuationOptions.NotOnFaulted | TaskContinuationOptions.ExecuteSynchronously,
					scheduler);

			return disposables;
		}

		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Reliability", "CA2000:Dispose objects before losing scope", Justification = "Returned to caller.")]
		public override IDisposable Schedule<TState>(TState state, Func<IScheduler, TState, IDisposable> action)
		{
			var cancel = new CancellationDisposable();
			var disposables = new CompositeDisposable(cancel);

			Task.Factory.StartNew(
				Run<TState>,
				Tuple.Create(action, this, state, disposables),
				cancel.Token,
				TaskCreationOptions.None,
				scheduler);

			return disposables;
		}

		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Reliability", "CA2000:Dispose objects before losing scope", Justification = "Returned to caller.")]
		public IDisposable ScheduleLongRunning<TState>(TState state, Action<TState, ICancelable> action)
		{
			var cancel = new CancellationDisposable();

			Task.Factory.StartNew(
				RunCancellable<TState>,
				Tuple.Create(action, state, (ICancelable) cancel),
				cancel.Token,
				TaskCreationOptions.LongRunning,
				scheduler);

			return cancel;
		}

		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Reliability", "CA2000:Dispose objects before losing scope", Justification = "Returned to caller.")]
		public IDisposable SchedulePeriodic<TState>(TState state, TimeSpan period, Func<TState, TState> action)
		{
			var cancel = new CancellationDisposable();
			var gate = new AsyncLock();

			TState localState = state;

			Action moveNext = null;
			moveNext = () =>
				Task.Delay(period, cancel.Token)
					.ContinueWith(
						_ =>
						{
							if (moveNext != null)
							{
								moveNext();
							}

							gate.Wait(() => localState = action(localState));
						},
					cancel.Token,
					TaskContinuationOptions.NotOnFaulted | TaskContinuationOptions.NotOnCanceled | TaskContinuationOptions.ExecuteSynchronously,
					scheduler);

			moveNext();

			return new CompositeDisposable(cancel, gate);
		}
	}
}