﻿using System;
using System.Collections.Generic;
using System.Reactive.Concurrency;
using System.Reactive.Linq;
using System.Reactive.Subjects;
using System.Threading;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Rxx.Parsers.Reactive.Linq;

namespace Rxx.UnitTests.Parsers.Reactive
{
	[TestClass]
	public class Projection : RxxTests
	{
		[TestMethod, TestCategory("Parsers"), TestCategory("Rx")]
		public void ParserSelectManyCollection_MovesToEndOfFirstMatch()
		{
			using (var parserCompleted = new ManualResetEventSlim())
			using (var source = new Subject<int>())
			{
				var results = source
					.ObserveOn(ThreadPoolScheduler.Instance)
					.Parse(parser =>
						from next in parser
						select from elements in next.Exactly(2)
									 from value in elements
									 where value < 2
									 select value);

				var values = new List<int>();

				using (results.Subscribe(values.Add, parserCompleted.Set))
				{
					source.OnNext(2);
					source.OnNext(2);

					source.OnNext(0);
					source.OnNext(1);

					source.OnNext(0);
					source.OnNext(1);

					source.OnCompleted();

					parserCompleted.Wait();

					Assert.AreEqual(4, values.Count);
				}
			}
		}

		[TestMethod, TestCategory("Parsers"), TestCategory("Rx")]
		public void ParserSelectManyCollection_MovesToEndOfFirstMatch_Async()
		{
			using (var artificialRaceCondition = new ManualResetEventSlim())
			using (var parserCompleted = new ManualResetEventSlim())
			using (var source = new Subject<int>())
			{
				var results = source
					.ObserveOn(ThreadPoolScheduler.Instance)
					.Parse(parser =>
						from next in parser
						select from elements in next.Exactly(2)
									 from value in elements
										.ObserveOn(CurrentThreadScheduler.Instance)
										.Do(_ => artificialRaceCondition.Wait())
									 where value < 2
									 select value);

				var values = new List<int>();

				using (results.Subscribe(values.Add, parserCompleted.Set))
				{
					source.OnNext(2);
					source.OnNext(2);

					source.OnNext(0);
					source.OnNext(1);

					source.OnNext(0);
					source.OnNext(1);

					source.OnCompleted();

					artificialRaceCondition.Set();

					parserCompleted.Wait();

					Assert.AreEqual(4, values.Count);
				}
			}
		}

		[TestMethod, TestCategory("Parsers"), TestCategory("Rx")]
		public void ParserSelectManyCollection_MovesToEndOfFirstMatch_Concurrent()
		{
			using (var artificialRaceCondition = new ManualResetEventSlim())
			using (var parserCompleted = new ManualResetEventSlim())
			using (var source = new Subject<int>())
			{
				var results = source
					.ObserveOn(ThreadPoolScheduler.Instance)
					.Parse(parser =>
						from next in parser
						select from elements in next.Exactly(2)
									 from value in elements
										.ObserveOn(ThreadPoolScheduler.Instance)
										.Do(_ => artificialRaceCondition.Wait())
									 where value < 2
									 select value);

				var values = new List<int>();

				using (results.Subscribe(values.Add, parserCompleted.Set))
				{
					source.OnNext(2);
					source.OnNext(2);

					source.OnNext(0);
					source.OnNext(1);

					source.OnNext(0);
					source.OnNext(1);

					source.OnCompleted();

					artificialRaceCondition.Set();

					parserCompleted.Wait();

					// TODO: Fix parsers so that this test always passes.  For now, we can't assert failure because sometimes the test passes.

					// Assert.AreEqual(4, values.Count);
				}
			}
		}

		[TestMethod, TestCategory("Parsers"), TestCategory("Rx")]
		public void ParserSelectManyCollection_MovesToEndOfFirstMatch_Nested()
		{
			using (var artificialRaceCondition = new ManualResetEventSlim())
			using (var parserCompleted = new ManualResetEventSlim())
			using (var source = new Subject<int>())
			{
				var results = source
					.ObserveOn(ThreadPoolScheduler.Instance)
					.Parse(parser =>
						from next in parser
						select from elements in next.Exactly(2)
									 from value in
										 elements.Parse(subparser =>
											 from subnext in subparser
											 select subnext.Where(
												 value =>
												 {
													 artificialRaceCondition.Wait();

													 return value < 2;
												 }))
									 select value);

				var values = new List<int>();

				using (results.Subscribe(values.Add, () => parserCompleted.Set()))
				{
					source.OnNext(2);
					source.OnNext(2);

					source.OnNext(0);
					source.OnNext(1);

					source.OnNext(0);
					source.OnNext(1);

					source.OnCompleted();

					artificialRaceCondition.Set();

					parserCompleted.Wait();

					Assert.AreEqual(4, values.Count);
				}
			}
		}
	}
}